/* File: AutoTester.java
 * Contains: A class for simulating Kqml input
 * Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Created: April 1998
 * Updated: Thu Mar  9 18:53:40 2000 by Jeff Dalton
 * Copyright: (c) 1998, AIAI, University of Edinburgh
 */

package ix.acp3;

import java.util.*;
import java.io.*;

import ix.util.*;
import ix.util.lisp.*;

/**
 * A class that can be used to replay recorded Kqml messages for testing
 * and demonstration.
 */

public class AutoTester extends Thread {

    protected static Vector allAutoTesters = new Vector();

    protected KqmlLogReader lin; 	// message source

    public int defaultDelay = 1000;	// milliseconds between messages
    public int maxDelay = 5000;		// milliseconds between messages

    public AutoTester(KqmlLogReader lin) {
	this.lin = lin;
	allAutoTesters.addElement(this);
    }

    public void run() {

// 	String line = Util.askLine("Type return to start:");
// 	if (line.equals("quit")) {
// 	    System.out.println("Ok, we won't run the test.\n");
// 	    return;
// 	}

	KqmlEvent event;
	Object message;
	long time = 0;

	while ((message = lin.nextMessage()) != Lisp.EOF) {
	    LList m = (LList)message;
	    long m_time = KqmlLogReader.messageDelta_t(m);
	    long delay;

	    if (m_time == 0)		// time not recorded?
		delay = defaultDelay;
	    else if (m_time > time) {
		delay = m_time - time;
		time = m_time;
	    }
	    else
		delay = 0;

	    if (delay > maxDelay) {
		Debug.noteln("Delay would be " + delay / 1000.0
			     + " seconds.");
		Debug.noteln("Using " + maxDelay / 1000.0
			     + " seconds instead.");
		delay = maxDelay;
	    }
	    
	    if (delay > 0) {
		try { Thread.sleep(delay); }
		catch (InterruptedException e) {}
	    }

	    event = new KqmlEvent(m);
	    Debug.noteln("Sending", event);
	    ACP3.watcher.interpret(event);
	}
    }

    /*
     * Reset
     */

    public static void stopAll() {
	for (Enumeration e = allAutoTesters.elements();
	     e.hasMoreElements();) {
	    AutoTester t = (AutoTester)e.nextElement();
	    if (t.isAlive())
		t.stop();
	}
    }

    /*
     * Test table, etc.
     */

    protected static Vector testNames = new Vector();

    protected static Hashtable testTable = new Hashtable() {
	public Object put(Object key, Object value) {
	    testNames.addElement(key); // to remember the order in which added
	    return super.put(key, value);
	}
    };

    static {
	testTable.put("Breach COA 1",       "log/acp3-breach-1-log");
	testTable.put("Breach COA 2, 3",    "log/acp3-breach-23-log");
	testTable.put("Breach COA 1, 2, 3", "log/acp3-breach-123-log");

	testTable.put("Test COA 1",         "log/breach-1/pcm.log");

//	testTable.put("Instant COA 1",      "log/acp3-instant-coa-1-log");

	testTable.put("Create finished COA 1", "log/acp3-instant-coa-1-log");

	testTable.put("Ask for a file name","");
    }

    public static Enumeration testNames() {
	return testNames.elements();
    }

    public static void runTest(final String testName) {
	// New thread because we're called from a MainFrame menu-slection
	// and hence in an AWT thread.
	new Thread() {
	    public void run() {
		Debug.noteln("Test requested", testName);
		String filename = (String)testTable.get(testName);
		if (filename == null)
		    Debug.warn("Unknown test: " + testName);
		else {
		    KqmlLogReader lr = filename.equals("")
			? askForFile()
			: getLogReader(filename);
		    if (lr != null)
			new AutoTester(lr).start();
		    else
			System.out.println("Ok, we won't run a test");
		}
	    }
	}.start();
    }

    protected static KqmlLogReader askForFile() {
	KqmlLogReader result = null;
	String fileName;
	System.out.println("\nType \"quit\" to exit.\n");
	while (result == null) {
	    fileName = Util.askLine("File to read:");
	    if (fileName.equals("quit"))
		break;
	    result = getLogReader(fileName);
	}
	return result;
    }

    protected static KqmlLogReader getLogReader(String fileName) {
	try {
	    return KqmlLogReader.newInstance(fileName);
	}
	catch (FileNotFoundException e) {
	    System.out.println("Can't find file \"" + fileName + "\"");
	}
	return null;
    }

}
