/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Tue Sep 12 17:28:16 2000 by Jeff Dalton
 * Copyright: (c) 2000, AIAI, University of Edinburgh
 */

package ix.iplan;

import java.io.*;
import ix.util.*;
import ix.util.lisp.*;
import ix.util.match.*;


public class LTF_Parser extends DomainParser implements LTF_SchemaSymbols {

    protected String filename;
    protected LispReader lin;

    public LTF_Parser(String filename) throws FileNotFoundException {
	this.filename = filename;
	this.lin = new LispFileReader(filename);
    }

    public Domain readDomain() {
	Debug.noteln("Reading schemas from", filename);
	Domain dom = new Domain();
	try {
	    Object item;
	    while ((item = lin.readObject()) != Lisp.EOF) {
		LList def = (LList)item;
		Symbol defForm = (Symbol)def.elementAt(0);
		if (defForm == S_DEFINE_PROCESS ||
		    defForm == S_DEFINE_ACTION) {
		    dom.addSchema(
		       new Schema(
			   canonicalSchemaDescription(def)));
		}
		else if (defForm == S_DEFINE_PRODUCT) {
		    dom.addProductSchema(
		       new ProductSchema(
			   canonicalProductDescription(def)));
		}
		else {
		    Debug.warn("Illegal definition type " + defForm);
		}
	    }
	}
	catch (LispReadException e) {
	    Debug.noteException(e);
	}
	return dom;	
    }

    /* Process and action definitions */

    public LList canonicalSchemaDescription(LList def) {
	Symbol defForm = (Symbol)def.elementAt(0);
	Symbol defType = null;
	Object defPat = def.elementAt(1);
	LList  defBody = def.drop(2);

	// Definition type: process or action
	if (defForm == S_DEFINE_PROCESS) {
	    defType = S_PROCESS;
	}
	else if (defForm == S_DEFINE_ACTION) {
	    defType = S_ACTION;
	}
	else {
	    Debug.warn("Illegal schema definition type" + defForm);
	}

	// Stick it all together as (type pattern . slot-alist)
	return Lisp.list(defType, defPat).append(parseBody(defBody));
    }

    public LList parseBody(LList body) {
	ListCollector result = new ListCollector();
	for (LList slots = body; slots != Lisp.NIL; slots = slots.cdr()) {
	    Cons slot = (Cons)slots.car();
	    Symbol name = (Symbol)slot.car();
	    LList value = slot.cdr();
	    if (name == S_NODES)
		result.addElement(new Cons(name, value));
	    else if (name == S_ORDERINGS)
		result.addElement(new Cons(name, value));
	    else if (name == S_PROPERTIES)
		result.addElement(new Cons(name, value));
	    else if (name == S_EFFECTS) {
		value = value.mapcar(new Function1() {
		    public Object funcall(Object description) {
			return parseEffect((LList)description);
		    }
		});
	        result.addElement(new Cons(name, value));
	    }
	    else
		Debug.warn("Illegal schema slot:" + slot);
	}
	return result.contents();
    }

    static final Object effectSyntax = Lisp.readFromString(
        "(at ?when ?pattern = ?value)"
    );

    static final Symbol
        Q_WHEN     = Symbol.intern("?when"),
	Q_PATTERN  = Symbol.intern("?pattern"),
	Q_VALUE    = Symbol.intern("?value");

    // /\/: Shouldn't have to repeat the "at ?when" part.  That is,
    // syntax should be (at ?when {?pattern = ?value}...).

    public static Effect parseEffect(LList description) {
	MatchEnv env = SimpleMatcher.mustMatch(effectSyntax, description);
	return new Effect(env.get(Q_WHEN),
			  env.get(Q_PATTERN),
			  env.get(Q_VALUE));
    }


    /* Product definitions */

    static Object defineProductSyntax = Lisp.readFromString(
      "(define-product (?name))"
    );

    static Symbol Q_NAME = Symbol.intern("?name");

    public LList canonicalProductDescription(LList def) {
	MatchEnv e = SimpleMatcher.mustMatch(defineProductSyntax, def);
	Object name = e.get(Q_NAME);
	return Lisp.list(Lisp.list(K_NAME, name));
    }

}

// Issues:
// * Add schemas one at a time or pass a list to new Domain()?
// * Tell the domain to analyse itself or leave it "raw"?
// * If we gave the domain a list of all schemas, it could
//   do the analysis automaticlly.  When we give it schemas
//   one at a time, it needs to be told it has them all.
// * Should new Schema() know how to precess an LList definition,
//   or should only the parser know?
// * Need to do some syntax-checking.
