/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Tue Sep 12 17:00:59 2000 by Jeff Dalton
 * Copyright: (c) 2000, AIAI, University of Edinburgh
 */

package ix.iplan;

import java.util.*;

import ix.util.*;
import ix.util.lisp.*;
import ix.util.match.*;

/**
 * A Schema describes a possible action in the process being modelled.
 * There can also be a schema that describes the top-level of the
 * process itself.
 */

public class Schema implements SchemaSymbols, Cloneable {
    public Symbol type;		// e.g. process or action
    public LList pattern;	// e.g. ("develop COA" ?coa)
    public Object action;	// e.g. "devalop COA"
    public LList nodes = Lisp.NIL;
    public LList orderings = Lisp.NIL;
    public LList properties = Lisp.NIL;
    public LList effects = Lisp.NIL;
    
    public boolean isForEachCoa; // true iff pattern contains ?coa /\/

    public Schema(LList def) {

	// def has the form (type pattern . slot-alist)

	this.type = (Symbol)def.elementAt(0);
	this.pattern = (LList)def.elementAt(1);
	this.action = pattern.elementAt(0);
	this.isForEachCoa = (pattern.elementAt(1) == Q_COA);

	Debug.noteln("Schema", pattern);

	// Init schema fields that correspond to slots in the definition.
	for (LList slots = def.drop(2); slots != Lisp.NIL; 
	     slots = slots.cdr()) {
	    Cons slot = (Cons)slots.car();
	    Symbol name = (Symbol)slot.car();
	    Object value = slot.cdr();
	    if (name == S_NODES)
		this.nodes = (LList)value;
	    else if (name == S_ORDERINGS)
		this.orderings = (LList)value;
	    else if (name == S_PROPERTIES)
		this.properties = (LList)value;
	    else if (name == S_EFFECTS)
		this.effects = (LList)value;
	    else
		Debug.warn("Illegal schema slot:" + slot);
	}
	// Some fields get additional processing
	this.orderings = expandAllOrderings(this.orderings);
	checkNodeNumbers();

    }

    private void checkNodeNumbers() {
	// Check that the nodes in the nodes clause are sequentally
	// numbered and start with 1.
	LList nodeSpecs;
	int place;
	for (nodeSpecs = nodes, place = 1;
	     nodeSpecs != Lisp.NIL; 
	     nodeSpecs = nodeSpecs.cdr(), place++) {
	    LList spec = (LList)nodeSpecs.car();
	    int number = ((Number)spec.car()).intValue();
	    if (number != place) {
		Debug.warn("Nodes not sequentially numbered from 1 " +
			   "in schema for " + pattern);
		break;
	    }
	}
	// Check that all orderings refer to node numbers that exist
	final int nodeMax = nodes.length();
	orderings.walkTree(new Function1() {
	    public Object funcall(Object nodeNumber) {
		int n = ((Number)nodeNumber).intValue();
		if (n < 1 || n > nodeMax)
		    Debug.warn("Ordering in schema for " + pattern +
			       "refers to nonexistent node " + nodeNumber);
		return null;
	    }
	});
    }

    public Object getPropertyObject (String name) {
	for (LList ps = properties; ps != Lisp.NIL; ps = ps.cdr()) {
	    Cons prop = (Cons)ps.car();
	    if (((String)prop.car()).equals(name)) {
		return prop.cdr().car();
	    }
	}
	return null;
    }

    public String getProperty(String name) {
	return (String)getPropertyObject(name);
    }

    public boolean isTrue(String name) {
	String value = getProperty(name);
	if (value == null || value.equals("false"))
	    return false;
	else {
	    Debug.assert(value.equals("true"), "not true, but not false");
	    return true;
	}
    }

    public Schema instantiate(final MatchEnv env) {
	try {
	    Schema s = (Schema)clone();
	    s.pattern = (LList)instantiateTree(s.pattern, env);
	    s.nodes = (LList)instantiateTree(s.nodes, env);
	    s.properties = (LList)instantiateTree(s.properties, env);
	    s.effects = s.effects.mapcar(new Function1() {
		public Object funcall(Object effect) {
		    return ((Effect)effect).instantiate(env);
		}
	    });
	    return s;
	}
	catch (CloneNotSupportedException e) {
	    Debug.noteException(e);
	    return null;
	}
    }

    public Schema forCoa(Long coa) {
	LList callPattern = isForEachCoa
	    ? Lisp.list(action, coa)
	    : Lisp.list(action);
	MatchEnv e = SimpleMatcher.match(pattern, callPattern);
	Debug.assert(e != null, "Can't instantiate schema for coa", pattern);
	return instantiate(e);
    }

    public LList toList() {			// for debugging
	Object[] form = new Object[] {
	    type, pattern,
		Lisp.list(S_NODES, nodes),
		Lisp.list(S_ORDERINGS, orderings),
		Lisp.list(S_PROPERTIES, properties)
		};
	return Seq.toList(Seq.elements(form));
    }

    public static Object instantiateTree(Object tree, MatchEnv env) {
	if (tree instanceof ItemVar) {
	    Object neu = env.get(tree);
	    Debug.assert(neu != null, 
			 "Tried to instantiate unbound var",
			 tree);
	    return neu;
	}
	else if (tree instanceof Cons) {
	    Cons c = (Cons)tree;
	    return new Cons(instantiateTree(c.car(), env),
			    (LList)instantiateTree(c.cdr(), env));
	}
	else
	    return tree;
    }

    LList expandAllOrderings(LList orderings) {
	ListCollector result = new ListCollector();
	for (LList ords = orderings; ords != Lisp.NIL; ords = ords.cdr()) {
	    result.concList(expandOrdering((LList)ords.car()));
	}
	return result.contents();
    }

    LList expandOrdering(LList ordering) {
	ListCollector result = new ListCollector();
	// for adjacent elements left and right ...
	for (LList ords = ordering; ords != Lisp.NIL; ords = ords.cdr()) {
	    Object left = ords.car();
	    Object right = ords.cdr().car();
	    result.concList(expandOrderPair(left, right));
	}
	return result.contents();
    }

    LList expandOrderPair(Object left, Object right) {
	ListCollector result = new ListCollector();
	for (LList l = ensureList(left); l != Lisp.NIL; l = l.cdr()) {
	    for (LList r = ensureList(right); r != Lisp.NIL; r = r.cdr()) {
		result.addElement(Lisp.list(l.car(), r.car()));
	    }
	}
	return result.contents();
    }

    LList ensureList(Object a) {
	return a instanceof LList ? (LList)a : Lisp.list(a);
    }

}

// Issues:
// * Property names should be either Keyword or Object, not String.
// * String property values should not be the default or have any
//   special status.
// * Ordering expansion should be done by the parser, because it's
//   syntax-dependent.
// * Should all methods that don't refer to the Schema be static?
// * Should there be so many public instance variables?
