/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Fri Sep 15 17:55:03 2000 by Jeff Dalton
 * Copyright: (c) 1998, AIAI, University of Edinburgh
 */

package ix.test;

import java.util.*;

import ix.icon.*;

import ix.util.*;
import ix.util.lisp.*;


/**
 * A simple example in an I-X framework.
 */

public class Trivial_IX_Test {

    /**
     * Main program.
     */
    public static void main(String[] argv) {

	T_IX_Component example = new T_IX_Component();

	example.addIssueHandlers(makeTestIssueHandlers());
	example.setListener(new T_IX_ComponentListener());
	example.start();

	example.newEvent(new T_Issue("twice", new T_Issue("return", "hello")));

	example.newEvent(new T_Issue("stop when finished"));

    }

    /**
     * Create issue handlers as "instant" subclasses of T_IssueHandler.
     */
    static Object[] makeTestIssueHandlers() {
	return new Object[] {

	    // Post an issue twice.
	    new T_IssueHandler("twice") {
		void handleIssue(T_Issue i) {
		    postIssue((Issue)i.object);
		    postIssue((Issue)i.object);
		}
	    },

            // Send something to the listener.
	    new T_IssueHandler("return") {
		void handleIssue(T_Issue i) {
		    component.notifyListener(i.object);
		}
	    },

	    // Stop the component.
	    new T_IssueHandler("stop when finished") {
		void handleIssue(T_Issue i) {
		    if (component.controller.issues.isEmpty())
			component.stop();
		    else
			postIssue(i); 	// go to the end of the queue
		}
	    }
	};
    }


    /**
     * Our issues contain a verb and an object.
     */
    static class T_Issue extends Issue {
	Object verb = Lisp.NIL;		// NIL's a printable null
	Object object = Lisp.NIL; 	// NIL's a printable null

	T_Issue(Object verb, Object object) {
	    this.verb = verb;
	    this.object = object;
	}

	T_Issue(Object verb) {
	    this.verb = verb;
	}

	public String toString() {
	    return "issue[" + Lisp.printToString(verb) + ", "
		            + Lisp.printToString(object) + "]";
	}
    }


    /**
     * A class that provides some basic structure, and some support
     * methods, for issue handlers.  Handlers for our example issues
     * are created by instantiating a subclass.
     */
    static abstract class T_IssueHandler extends IssueHandler {
	Object verb;
	T_IX_Component component;

	T_IssueHandler(Object verb) {
	    this.verb = verb;
	}

	void setComponent(T_IX_Component component) {
	    this.component = component;
	}

	abstract void handleIssue(T_Issue i);

	void postIssue(Issue i) {
	    component.controller.addIssue(i);
	}

    }


    /**
     * Information is sent out via a listener.
     */
    static class T_IX_ComponentListener {

	public T_IX_ComponentListener() {}

	void receive(Object message) {
	    Debug.noteln("Listener received:", message);
	}

    }


    /**
     * A simple I-X entity with its own thread.
     */
    static class T_IX_Component implements Runnable {
	Thread thread;
	MessageQueue q = new MessageQueue();
	T_IX_ComponentListener listener;
	T_Controller controller;

	T_IX_Component() { }

	void addIssueHandlers(Object[] IssueHandlers) {
	    controller = new T_Controller(this);
	    controller.installHandlers(IssueHandlers);
	    thread = new Thread(this);
	}

	public void setListener(T_IX_ComponentListener listener) {
	    this.listener = listener;
	}

	void notifyListener(Object message) {
	    if (listener != null)
		listener.receive(message);
	    else
		Debug.noteln("No listener for", message);
	}
	
	public void start() {
	    thread.start();
	}

	public void stop() {
	    thread.stop();
	}

	public void newEvent(Object e) {
	    q.send(e);
	}

	Issue eventToIssue(Object e) {
	    return (Issue)e;
	}

	public void run() {
	    controller.mainLoop();
	}

    }


    /**
     * A very simple controller.
     */
    static class T_Controller extends Controller {
	T_IX_Component component;
	MessageQueue q;
	ListCollector issues = new ListCollector();
	Hashtable handlerTable = new Hashtable();

	T_Controller(T_IX_Component component) {
	     this.component = component;
	     this.q = component.q;
	}

	void mainLoop() {
	    // Do this forever.
	    while (true) {
		// Convert any events to Issues.
		while (q.hasMessage()) {
		    addEventIssue(q);
		}
		// If there's an issue, handle it;
		// otherwise wait for events.
		if (!issues.isEmpty()) {
		    handleIssue(selectIssue());
		}
		else {
		    addEventIssue(q);
		}
	    }
	}

	protected void addIssue(Issue i) {
	    Debug.noteln("Adding", i);
	    issues.addElement(i);
	}

	protected void addEventIssue(MessageQueue q) {
	    // N.B. the component translates the event to an issue.
	    addIssue(component.eventToIssue(q.nextMessage()));
	}

	Issue selectIssue() {
	    return (Issue)issues.popElement();
	}

	void installHandlers(Object[] IssueHandlers) {
	    // Put the knowledge sources into the handler table
	    // and tell each of them about their controller.
	    for (Enumeration ke = Seq.elements(IssueHandlers);
		 ke.hasMoreElements();) {
		T_IssueHandler k = (T_IssueHandler)ke.nextElement();
		k.setComponent(component);
		handlerTable.put(k.verb, k);
	    }
	}

	T_IssueHandler findHandler(T_Issue i) {
	    T_IssueHandler k = (T_IssueHandler)handlerTable.get(i.verb);
	    Debug.assert(k != null, "no KS for", i.verb);
	    return k;
	}

	void handleIssue(Issue i) {
	    T_Issue ti = (T_Issue)i;
	    Debug.noteln("Handling", ti);
	    findHandler(ti).handleIssue(ti);
	}

    }

}

