/* Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Updated: Mon Nov 27 16:47:20 2000 by Jeff Dalton
 * Copyright: (c) 2000, AIAI, University of Edinburgh
 */

package ix.examples;

import java.util.*;

import java.io.FileNotFoundException;

import ix.util.*;
import ix.util.lisp.*;

import ix.iplan.*;


// /\/: Straighten out what should be public, protected, etc.


/**
 * A simple example using an I-X framework.  This example shows how
 * to extend PicoIX for a particular application type, in this case
 * discrete event simulation.
 */

public class MicroISim extends PicoISim {

    static final Symbol
        S_NEW            = Symbol.intern("new"),
	S_WAITING        = Symbol.intern("waiting"),
        S_READY          = Symbol.intern("ready"),
	S_EXECUTING      = Symbol.intern("executing"),
	S_COMPLETE       = Symbol.intern("complete");

    static Symbol SYM(String name) {		// Abbreviation
	return Symbol.intern(name);
    }


    /**
     * An IX_System for discrete event simulation.
     */
    static class IX_Simulator extends IX_System {

	long simTime = -1;	// so it notes a change to 0

	boolean stopWhenIdle = true;

	Domain domain = null;

	IX_Simulator() {
	    controller =  new SimController(this);
	    model = new ModelManager(this);
	    addIssueHandlers(makeBuiltinIssueHandlers());
	    addSimEventHandlers(makeBuiltinSimEventHandlers());
	}

	IX_Simulator(Controller c, ModelManager m) {
	    super(c != null ? c : new SimController(),
		  m);
	    addIssueHandlers(makeBuiltinIssueHandlers());
	    addSimEventHandlers(makeBuiltinSimEventHandlers());
	}

	void addSimEventHandlers(Object[] issueHandlers) {
	    ((SimController)controller).installSimEventHandlers(issueHandlers);
	}

	void setstopWhenIdle(boolean newValue) {
	    stopWhenIdle = newValue;
	}

	void setDomain(Domain domain) {
	    this.domain = domain;
	}

	void setDomain(String filename) throws FileNotFoundException {
	    domain = DomainParser.makeParser(filename).readDomain();
	}

	void schedule(SimEvent e) {
	    newEvent(new Issue(SYM("do"), e));
	}

// 	void schedule(Object actionPattern, long when) {
// 	    newEvent(new Issue(SYM("do"), 
// 			       new Issue_Execute(actionPattern, when)));
// 	}

	void setSimTime(long t) {
	    if (t > simTime) {
		Debug.noteln("\n- - - - - time now " + t + " - - - - -\n");
		simTime = t;
	    }
	}

	void postSimEvent(SimEvent e) {
	    ((SimController)controller).addSimEvent(e);
	}

    }

    static Object[] makeBuiltinIssueHandlers() {
	return new Object[] {

	    new IssueHandler(SYM("do")) {
		void handleIssue(Issue i) {
		    SimEvent e = (SimEvent)i.object;
		    ((IX_Simulator)system).postSimEvent(e);
		}
	    }

	};

    }

    static Object[] makeBuiltinSimEventHandlers() {
	return new Object[] {

	    new KS_Execute()

	};

    }


    /**
     * The controller for a simulator handles an agenda of simulated
     * events as well as the agenda of issues that the simulator addresses
     * as an agent.
     */
    static class SimController extends Controller {

	ListCollector simEvents = new ListCollector();
	Hashtable simHandlerTable = new Hashtable();

	SimController() {}

	SimController(IX_System system) {
	    super(system);
	}

	void addSimEvent(SimEvent e) {
	    Debug.noteln("Scheduling", e);
	    e.computeStatus();
	    simEvents.insertElement(e, EARLIER_EST_P);
	}

	static final Predicate2 EARLIER_EST_P =
	    new Predicate2() {
		public boolean trueOf(Object a, Object b) {
		    return ((SimEvent)a).est < ((SimEvent)b).est;
		}
	    };

	void mainLoop() {
	    // Do this forever.
	    while (true) {
		// Convert any external events to Issues.
		while (q.hasMessage()) {
		    addEventIssue(q);
		}
		// If there's an issue, handle it;
		// else if we can execute a SimEvent, do that;
		// otherwise wait for an external event
		// unless we're supposed to stop when idle.
		if (!issues.isEmpty()) {
		    handleIssue(selectIssue());
		}
		else if (!simEvents.isEmpty()) {
		    boolean didSomething = simulateIfPossible();
		    // If we get here and didn't do anything, we'll keep
		    // being unable to do anything until some external
		    // event makes a change that lets us progress,
		    // so we may as well wait for an event rather than
		    // keep looping.
		    if (!didSomething) {
			addEventIssue(q); 	// waits
		    }
		}
		else if (((IX_Simulator)system).stopWhenIdle) {
		    system.stop();
		    Debug.assert(false, "continued after thread stop()");
		}
		else {
		    addEventIssue(q); 		// waits
		}
	    }
	}

	boolean simulateIfPossible() {
	    Debug.noteln("Agenda", simEvents.contents());
	    for(LList el = simEvents.contents();
		el != Lisp.NIL; el = el.cdr()) {
		SimEvent e = (SimEvent)el.car();
		if (canSimulate(e)) {
		    simEvents.deleteElement(e);
		    setSimTime(e.est);
		    simulate(e);
		    return true;
		}
	    }
	    return false;
	}

	boolean canSimulate(SimEvent e) {
	    SimEventHandler h = findSimHandler(e);
	    return e.status == S_READY && h.isAble(e);
	}

	void simulate(SimEvent e) {
	    SimEventHandler h = findSimHandler(e);
	    h.handleIssue(e);
	}

	void installSimEventHandlers(Object[] issueHandlers) {
	    installIssueHandlers(issueHandlers, simHandlerTable);
	}

	SimEventHandler findSimHandler(Issue i) {
	    SimEventHandler h = (SimEventHandler)simHandlerTable.get(i.verb);
	    Debug.assert(h != null, "no sim handler for", i.verb);
	    return h;
	}

	// It's the controller that sets the sim time because different
	// controllers might handle simulated time in different ways,
	// for instance by using scaled real time.
	void setSimTime(long t) {
	    ((IX_Simulator)system).setSimTime(t);
	}

    }


    /**
     * Issues that represent scheduled events in the simulation.
     */
    static abstract class SimEvent extends Issue {

	long est = 0;
	Symbol status = S_NEW;
	SimEvent parent = null;
	LList children = Lisp.NIL;
	LList predecessors = Lisp.NIL;
	LList successors = Lisp.NIL;

	SimEvent(Object verb, Object object, long when) {
	    super(verb, object);
	    this.est = when;
	}

	void execute(IX_Simulator simulator) {
	    // The default does nothing.
	    // The simulator is passed as an argument to allow the
	    // method to access other parts of the system.
	}

	Symbol getStatus() {
	    return status;
	}

	void setStatus(Symbol newStatus) {
	    status = newStatus;
	    // If the new status is COMPLETE, that may affect other events.
	    if (status == S_COMPLETE) {
		// parent might become COMPLETE
		if (parent != null)
		    parent.computeStatus();
		// a successor might become READY
		for (Enumeration se = successors.elements();
		     se.hasMoreElements();) {
		    SimEvent s = (SimEvent)se.nextElement();
		    s.computeStatus();
		}
	    }
	}

	void computeStatus() {
	    // If status is NEW, start by assuming WAITING instead.
	    if (status == S_NEW) {
		status = S_WAITING;
	    }
	    // Check for standard automatic transitions
	    if (status == S_WAITING) {
		// WAITING - may become READY
		if (allHaveStatus(predecessors, S_COMPLETE)) {
		    setStatus(S_READY);
		}
	    }
	    else if (status == S_EXECUTING) {
		// EXECUTING - may become COMPLETE
		Debug.assert(children != Lisp.NIL,
			     "executing non-parent", this);
		if (allHaveStatus(children, S_COMPLETE)) {
		    setStatus(S_COMPLETE);
		}
	    }
	}

	static boolean allHaveStatus(LList events, Symbol status) {
	    for (Enumeration ee = events.elements();
		 ee.hasMoreElements(); ) {
		SimEvent e = (SimEvent)ee.nextElement();
		if (e.getStatus() != status)
		    return false;
	    }
	    return true;
	}

	public String toString() {
	    return "SimEvent[" + Lisp.printToString(verb) + " "
		               + Lisp.printToString(object) + " at "
		               + est + "]";
	}

    }


    /**
     * Handlers for SimEvents.
     */
    static abstract class SimEventHandler extends IssueHandler {

	SimEventHandler(Object verb) {
	    super(verb);
	}

	// We distinguish between being ready to execute and being able.
	// For the latter, we need at least the issue handler's help.
	abstract boolean isAble(SimEvent e);

	// Handy methods for use in SimEvent handlers.

	long getSimTime() {
	    return ((IX_Simulator)system).simTime;
	}

	void postSimEvent(SimEvent e) {
	    ((IX_Simulator)system).postSimEvent(e);
	}

    }


    /**
     * "execute" events
     */
    static class Issue_Execute extends SimEvent {

	LList schemas = null;
	LList filteredSchemas = null;

	Issue_Execute(Object actionPattern, long when) {
	    super(SYM("execute"), actionPattern, when);
	}

	void ensureSchemas(Domain dom) {
	    if (schemas == null) {
		if (object instanceof LList && dom != null) {
		    schemas = dom.getMatchingSchemas((LList)object);
		}
		else {
		    // No valid pattern, so don't even ask for schemas.
		    schemas = Lisp.NIL;
		}
	    }
	}

	void filterSchemas() {
	    filteredSchemas = schemas; 	// for now /\/
	}

    }



    /**
     * The handler for "execute" events.
     */
    static class KS_Execute extends SimEventHandler {

	KS_Execute() {
	    super(SYM("execute"));
	}

	KS_Execute(Object verb) {		// for instant subclasses
	    super(verb);
	}

	void handleIssue(Issue i) {
	    // We can assume isAble() has just been called on the same issue.
	    SimEvent e = (SimEvent)i;

	    Debug.noteln("Executing", e);
	    e.setStatus(S_EXECUTING);

	    // ... Execution ...
	    e.execute((IX_Simulator)system);

	    if (e.children.isNull())
		e.setStatus(S_COMPLETE);

	}

	boolean isAble(SimEvent event) {
	    Issue_Execute e = (Issue_Execute)event;
	    Domain dom = ((IX_Simulator)system).domain;

	    e.ensureSchemas(dom);
	    if (e.schemas.isNull()) {
		return true;
	    }
	    else {
		e.filterSchemas();
		return !e.filteredSchemas.isNull();
	    }

	}

	boolean checkIssuePreconditions(LList preconditions) {
	    for (Enumeration pe = preconditions.elements();
		 pe.hasMoreElements();) {
		Constraint c = (Constraint)pe.nextElement();
		if (!tryConstraint(c).equals(Boolean.TRUE))
		    return false;
	    }
	    return true;
	}

    }


    /**
     * Constraints for a "pattern = value" world state.
     */
    static class WorldStateConstraint extends Constraint {

	// An Object for matching these constraints with the CM.
	static final Symbol S_WORLD = Symbol.intern("world");

	// Constraint subtypes
	static final Symbol
	    S_CONDITION = Symbol.intern("condition"),
	    S_EFFECT    = Symbol.intern("effect");

	Object subtype;
	Object pattern;
	Object value;

	WorldStateConstraint(Object subtype, Object pattern, Object value) {
	    super(S_WORLD);			// type
	    this.subtype = subtype;
	    this.pattern = pattern;
	    this.value = value;
	    this.args = Lisp.list(subtype, pattern, value); // for debugging
	}

    }


    /**
     * A constraint manager for a "pattern = value" world state.
     */
    static class WorldStateCM extends ConstraintManager {

	Object type = WorldStateConstraint.S_WORLD; 	// what we manage

	Hashtable stateTable = new Hashtable();

        WorldStateCM() {
	    super(WorldStateConstraint.S_WORLD); 	// type
	}

	Object addConstraint(Constraint c) {
	    WorldStateConstraint wsc = (WorldStateConstraint)c;
	    if (wsc.subtype == WorldStateConstraint.S_CONDITION) {
		return addCondition(wsc);
	    }
	    else {
		Debug.assert(wsc.subtype == WorldStateConstraint.S_EFFECT);
		return addEffect(wsc);
	    }
	}

	protected Object addCondition(WorldStateConstraint cond) {
	    return tryCondition(cond);
	}

	protected Object addEffect(WorldStateConstraint effect) {
	    stateTable.put(effect.pattern, effect.value);
	    return Boolean.TRUE;
	}

	Object tryConstraint(Constraint c) {
	    WorldStateConstraint wsc = (WorldStateConstraint)c;
	    if (wsc.subtype == WorldStateConstraint.S_CONDITION) {
		return tryCondition(wsc);
	    }
	    else {
		Debug.assert(wsc.subtype == WorldStateConstraint.S_EFFECT);
		return Boolean.TRUE;
	    }
	}

 	protected Object tryCondition(WorldStateConstraint cond) {
 	    Object value = stateTable.get(cond.pattern);
 	    return new Boolean(value != null && value.equals(cond.value));
 	}

    }


}

// Issues:
// * Division of labour between Issue and IssueHandler methods.
// * Could the handler just be a method of the Issue?
// * Should isAble(SimEvent s) be a handler method, or should
//   we have isAble(SimEventHandler h) be a method of SimEvent?
//   (The handler is involved because it knows that the system is.)
// * Instead of having the world state constraint and CM classes here,
//   it might be better to have a separate top-level SimWorldState
//   class containing static subclasses Constraint and CM?  Users would
//   write SimWorldState.Constraint and SimWorldState.CM.
//   Issue_Execute and KS_Execute might be done the same way.
// * Should SimEvent implement Runnable?  Don't want to imply that
//   all will work unproblematically if they're run in a separate
//   thread.
