/* File: ViewMatrix.java
 * Contains: IP2 ViewMatrix
 * Author: Austin Tate <a.tate@ed.ac.uk>, Jeff Dalton
 * Created: Fri Jan 16 11:13:15 1998
 * Updated: Thu Jan 18 14:34:30 2001 by Jeff Dalton
 * Copyright: (c) 1998, 2000, AIAI, University of Edinburgh
 */

package ix.view.matrix;

import java.awt.*;
import java.awt.image.*;
import java.io.File;
import java.util.*;

import ix.ip2.*;
import ix.util.*;
import ix.util.lisp.*;

public class ViewMatrix extends MatrixPanel {

    // A tree-as-LList description of the process
    // N.B. This is not cleared by a reset.  /\/
    LList processExpansionTree = null;
                
    public ViewMatrix(Container parent) {
        super(parent);
    }

    public void defaultState () {
	if (processExpansionTree == null) {
	    Debug.noteln("ViewMatrix: asking for expansion tree");
	    processExpansionTree = IP2.model.staticExpansionTree();
	}
	loadExpansionTree(processExpansionTree);
    }

    public void loadExpansionTree(LList tree) {
	Vector labels = new Vector();
	Vector levels = new Vector();

	Debug.noteln("ViewMatrix: Loading expansion tree " + tree);

	processExpansionTree = tree; 	// remember /\/

	labels.addElement("Issues Remaining");
	levels.addElement(new Integer(0));
	vectorizeExpanstionTree(tree, 0, labels, levels);
	Debug.assert(labels.size() == levels.size());

	String[] pLabel = new String[labels.size()];
	int[] pLevel = new int[levels.size()];
	for (int i = 0; i < pLabel.length; i++) {
	    pLabel[i] = labels.elementAt(i).toString();
	    pLevel[i] = ((Integer)levels.elementAt(i)).intValue();
	}

	loadMatrix(pLabel, pLevel, false, true, "   Overall   ", "COA-");

	// setNumberOfCOA(1);

    }

    // The expansion tree shows the action nesting present in the plan.
    // The tree is a list of entries of the form (pattern subtree).

    protected void vectorizeExpanstionTree(LList tree, int depth, 
					   Vector labels, Vector levels) {
	for (LList at = tree; at != Lisp.NIL; at = at.cdr()) {
	    LList e = (LList)at.car();
	    Object pattern = (Object)e.elementAt(0);
	    LList subtree = (LList)e.elementAt(1);
	    labels.addElement(pattern);
	    levels.addElement(new Integer(depth));
	    vectorizeExpanstionTree(subtree, depth+1, labels, levels);
	}
    }

    protected void handleStatusUpdate(ProcessStatusUpdate up) {

	setDrawOn(false);

	if (up.containsCompleteState) {
	    // Setting the number of COA to 0 will have the side-effect
	    // of resetting all the status values to BLANK.  This isn't
	    // really necessary (modulo bugs), except for an empty model,
	    // because the update contains a complete description.
	    // setNumberOfCOA doesn't set status in the overall column.  /\/
	    setNumberOfCOA(0);
	    // If the model's empty, make sure the overall column is 
	    // initialized.
	    if (up.statusChanges.isEmpty()) {
		// First set everything in the overall column to BLANK.
		for (int i = 1; i <= maxRows; i++)
		    setStatus(i,0,ViewColor.STATUS_BLANK,null);
		// The top entry in the overall column is POSSIBLE.
		// The process model can't easily tell us this, because
		// it doesn't know enough until there's a COA.  /\/
		// -- setStatus(1, 0, ViewColor.STATUS_POSSIBLE, null);
		// Now there's one issue in the overall column.
		setIssue(0,ViewColor.STATUS_NA,"1");
	    }
	    // Start with nothing expanded.
	    resetToContracted();
	    // Expand some things
	    // setExpanded(getLabelIndex("develop COA"));
	    // setExpanded(getLabelIndex("create COA"));
	    // setExpanded(getLabelIndex("create support plan"));
	}

	setNumberOfCOA(up.numberOfCoas);

	// Node status changes.
	for (Enumeration ce = up.statusChanges.elements();
	     ce.hasMoreElements();) {
	    StatusChange change = (StatusChange)ce.nextElement();
	    int i = getLabelIndex(change.action);
	    String text = change.statusText();
	    if (change.coa >= 1) {
		setStatus(i, change.coa, change.status, text);
	    }
	    else if (change.coa == StatusChange.COA_OVERALL)
		setStatus(i, 0, change.status, text);
	    else {
		Debug.assert(change.coa == StatusChange.COA_NONE);
		// Put the status value in every entry on the row.
		for (int coa = 1; coa <= up.numberOfCoas; coa++) {
		    setStatus(i, coa, change.status, text);
		}
	    }
	}

	// Issue counts
	Debug.assert(up.issueCounts.length == up.numberOfCoas + 1);
	for (int coa = 0; coa < up.issueCounts.length; coa++)
	    setIssue(coa,
		     ViewColor.STATUS_NA,
		     String.valueOf(up.issueCounts[coa]));

	setDrawOn(true);

    }

}
  
